from typing import List, Optional, Dict, Set, Any

from pydantic import BaseModel, validator

from unipipeline.modules.uni_broker_definition import UniBrokerDefinition
from unipipeline.modules.uni_message_type_definition import UniMessageTypeDefinition
from unipipeline.modules.uni_module_definition import UniModuleDefinition
from unipipeline.modules.uni_waiting_definition import UniWaitingDefinition

WORKER_ERROR_NAME = '__error__'


class UniWorkerDefinition(BaseModel):
    name: str
    broker: UniBrokerDefinition[Any]
    type: UniModuleDefinition
    topic: str
    prefetch: int
    input_message: UniMessageTypeDefinition
    output_workers: List[str]
    retry_max_count: int
    retry_delay_s: int
    max_ttl_s: Optional[int]
    auto_ack: bool
    waitings: List[UniWaitingDefinition]

    @validator("prefetch")
    def prefetch_must_be_gte_1(cls, v: int) -> Optional[int]:
        if v < 1:
            raise ValueError("must be >= 1")
        return v

    @validator("retry_max_count")
    def retry_max_count_must_be_gt_0(cls, v: int) -> Optional[int]:
        if v < 0:
            raise ValueError("must be >= 0")
        return v

    @validator("retry_delay_s")
    def retry_delay_s_must_be_gt_0(cls, v: int) -> Optional[int]:
        if v < 0:
            raise ValueError("must be >= 0")
        return v

    def get_related_broker_names(self, index_of_workers: Dict[str, 'UniWorkerDefinition']) -> Set[str]:
        result = {self.broker.name, index_of_workers[WORKER_ERROR_NAME].broker.name}
        for w_name in self.output_workers:
            result.add(index_of_workers[w_name].broker.name)
        return result

    def get_related_worker_names(self) -> Set[str]:
        result = {self.name, WORKER_ERROR_NAME}
        for w_name in self.output_workers:
            result.add(w_name)
        return result

    def wait_everything(self) -> None:
        for wd in self.waitings:
            wd.wait()



import logging
from time import sleep
from typing import Callable, TypeVar, Any

from unipipeline.modules.uni_broker_definition import UniBrokerDefinition
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_message_meta import UniMessageMeta

TMessage = TypeVar('TMessage', bound=UniMessage)
logger = logging.getLogger(__name__)


class UniBrokerMessageManager:
    def reject(self) -> None:
        raise NotImplementedError(f'method reject must be specified for class "{type(self).__name__}"')

    def ack(self) -> None:
        raise NotImplementedError(f'method acknowledge must be specified for class "{type(self).__name__}"')


class UniBroker:
    def __init__(self, definition: UniBrokerDefinition[Any]) -> None:
        self._definition = definition

    def connect(self) -> None:
        raise NotImplementedError(f'method connect must be implemented for {type(self).__name__}')

    def close(self) -> None:
        raise NotImplementedError(f'method close must be implemented for {type(self).__name__}')

    def consume(
        self,
        topic: str,
        processor: Callable[[UniMessageMeta, UniBrokerMessageManager], None],
        consumer_tag: str,
        worker_name: str,
        prefetch: int = 1
    ) -> None:
        raise NotImplementedError(f'method consume must be implemented for {type(self).__name__}')

    def publish(self, topic: str, meta: UniMessageMeta) -> None:
        raise NotImplementedError(f'method consume must be implemented for {type(self).__name__}')

    @property
    def definition(self) -> UniBrokerDefinition[Any]:
        return self._definition

    @staticmethod
    def waiting_for_connection(definition: UniBrokerDefinition[Any]) -> 'UniBroker':
        broker_type = definition.type.import_class(UniBroker)
        for try_count in range(definition.retry_max_count):
            try:
                b = broker_type(definition=definition)
                b.connect()
                logger.debug('%s is available', broker_type.__name__)
                return b
            except Exception as e:
                logger.debug('retry connect to broker %s [%s/%s] : %s', broker_type.__name__, try_count, definition.retry_max_count, str(e))
                sleep(definition.retry_delay_s)
                continue
        raise ConnectionError(f'unavailable connection to {broker_type.__name__}')



from typing import NamedTuple

from unipipeline.modules.uni_module_definition import UniModuleDefinition


class UniMessageTypeDefinition(NamedTuple):
    name: str
    type: UniModuleDefinition
    version: int



from importlib import import_module
from typing import NamedTuple, Generic, Type, TypeVar

T = TypeVar('T')


class UniModuleDefinition(NamedTuple, Generic[T]):
    module: str
    class_name: str

    def import_class(self, type: Type[T]) -> Type[T]:
        mdl = import_module(self.module)
        tp = getattr(mdl, self.class_name)
        assert issubclass(tp, type)
        return tp



from unipipeline.modules.uni_module_definition import UniModuleDefinition


def parse_type(type_def: str) -> UniModuleDefinition:
    assert isinstance(type_def, str), f"type_def must be str. {type(type_def)} given"
    spec = type_def.split(":")
    assert len(spec) == 2, f'must have 2 segments. {len(spec)} was given from "{type_def}"'
    return UniModuleDefinition(
        module=spec[0],
        class_name=spec[1],
    )
